----------------------------------------------------------------------------------------------------
-- PipeBridge
----------------------------------------------------------------------------------------------------
-- Purpose: Link pipe from output to input vehicles.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class PipeBridge
PipeBridge = {}
PipeBridge.MOD_NAME = g_currentModName

function PipeBridge.prerequisitesPresent(specializations)
    return true
end

function PipeBridge.initSpecialization()
    g_configurationManager:addConfigurationType("pipeBridge", g_i18n:getText("configuration_pipeBridge"), nil, nil, nil, nil, ConfigurationUtil.SELECTOR_MULTIOPTION)
end

function PipeBridge.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "loadPipesFromXML", PipeBridge.loadPipesFromXML)
    SpecializationUtil.registerFunction(vehicleType, "loadOutputPipe", PipeBridge.loadOutputPipe)
    SpecializationUtil.registerFunction(vehicleType, "loadInputPipe", PipeBridge.loadInputPipe)
    SpecializationUtil.registerFunction(vehicleType, "getOutputPipes", PipeBridge.getOutputPipes)
    SpecializationUtil.registerFunction(vehicleType, "hasOutputPipes", PipeBridge.hasOutputPipes)
    SpecializationUtil.registerFunction(vehicleType, "getInputPipes", PipeBridge.getInputPipes)
    SpecializationUtil.registerFunction(vehicleType, "hasInputPipes", PipeBridge.hasInputPipes)
    SpecializationUtil.registerFunction(vehicleType, "bridgePipesWithVehicle", PipeBridge.bridgePipesWithVehicle)
    SpecializationUtil.registerFunction(vehicleType, "canBridgePipesWithVehicle", PipeBridge.canBridgePipesWithVehicle)
    SpecializationUtil.registerFunction(vehicleType, "unlinkPipesWithVehicle", PipeBridge.unlinkPipesWithVehicle)
    SpecializationUtil.registerFunction(vehicleType, "getPipeBridgeAttacher", PipeBridge.getPipeBridgeAttacher)
    SpecializationUtil.registerFunction(vehicleType, "updateAttacherJoint", PipeBridge.updateAttacherJoint)
    SpecializationUtil.registerFunction(vehicleType, "setAttacherJointHeight", PipeBridge.setAttacherJointHeight)
end

function PipeBridge.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateInterpolation", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onPostAttach", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onPreDetach", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onReadStream", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteStream", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", PipeBridge)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", PipeBridge)
end

function PipeBridge.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "isDetachAllowed", PipeBridge.isDetachAllowed)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "canTurnOnPump", PipeBridge.canTurnOnPump)
end

function PipeBridge:onLoad(savegame)
    self.spec_pipeBridge = self[("spec_%s.pipeBridge"):format(PipeBridge.MOD_NAME)]
    local spec = self.spec_pipeBridge

    local configurationId = Utils.getNoNil(self.configurations["pipeBridge"], 1)
    local baseKey = ("vehicle.pipeBridgeConfigurations.pipeBridgeConfiguration(%d)"):format(configurationId - 1)
    ObjectChangeUtil.updateObjectChanges(self.xmlFile, "vehicle.pipeBridgeConfigurations.pipeBridgeConfiguration", configurationId, self.components, self)

    --Fallback
    if not hasXMLProperty(self.xmlFile, baseKey) then
        baseKey = "vehicle"
    end

    spec.outputPipes = self:loadPipesFromXML(self.xmlFile, baseKey .. ".pipeBridge.output", self.loadOutputPipe)
    spec.inputPipes = self:loadPipesFromXML(self.xmlFile, baseKey .. ".pipeBridge.input", self.loadInputPipe)

    -- Root node is used to re-link outputs on detach.
    spec.rootNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. ".pipeBridge#rootNode"), self.i3dMappings)

    spec.forceHigherOnAttach = Utils.getNoNil(getXMLBool(self.xmlFile, baseKey .. ".pipeBridge#forceHigher"), false)
    spec.numOfAttachedImplements = 0
    spec.numOfAttachedImplementsSent = 0
    spec.attacherVehicle = nil
    spec.moveAlpha = 1
    spec.dirtyFlag = self:getNextDirtyFlag()
end

function PipeBridge:onDelete()
    local spec = self.spec_pipeBridge
    if spec.tanker ~= nil then
        local hoses = self:getConnectionHosesByInputAttacherJoint(self:getActiveInputAttacherJointDescIndex())
        for _, hose in ipairs(hoses) do
            self:disconnectHose(hose)
        end
        spec.tanker = nil
    end
end

function PipeBridge:onReadStream(streamId, connection)
    if connection:getIsServer() then
        local spec = self.spec_pipeBridge
        if streamReadBool(streamId) then
            local moveAlpha = streamReadFloat32(streamId)
            self:updateAttacherJoint(moveAlpha)
        end

        spec.numOfAttachedImplementsSent = streamReadInt8(streamId)
    end
end

function PipeBridge:onWriteStream(streamId, connection)
    if not connection:getIsServer() then
        local spec = self.spec_pipeBridge
        if streamWriteBool(streamId, spec.jointDesc ~= nil) then
            streamWriteFloat32(streamId, spec.moveAlpha)
        end

        streamWriteInt8(streamId, spec.numOfAttachedImplementsSent)
    end
end

function PipeBridge:onReadUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() and streamReadBool(streamId) then
        local spec = self.spec_pipeBridge
        local moveAlpha = streamReadFloat32(streamId)
        self:updateAttacherJoint(moveAlpha)
        spec.numOfAttachedImplementsSent = streamReadInt8(streamId)
    end
end

function PipeBridge:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        local spec = self.spec_pipeBridge
        if streamWriteBool(streamId, bitAND(dirtyMask, spec.dirtyFlag) ~= 0) then
            streamWriteFloat32(streamId, spec.moveAlpha)
            streamWriteInt8(streamId, spec.numOfAttachedImplementsSent)
        end
    end
end

function PipeBridge:onUpdate(dt)
    local spec = self.spec_pipeBridge

    if not self.isServer then
        return
    end

    if spec.jointDesc ~= nil then
        if spec.heightTargetAlpha ~= -1 then
            local diff = spec.heightTargetAlpha - spec.moveAlpha + 0.0001
            local moveTime = diff / (spec.jointDesc.upperAlpha - spec.jointDesc.lowerAlpha) * spec.jointDesc.moveTime
            local moveStep = dt / moveTime * diff

            if diff > 0 then
                moveStep = -moveStep
            end

            local newAlpha = spec.moveAlpha + moveStep
            self:updateAttacherJoint(newAlpha)

            if math.abs(spec.heightTargetAlpha - newAlpha) < 0.01 then
                spec.heightTargetAlpha = -1
            end
        end
    end
end

function PipeBridge:onUpdateInterpolation(dt)
    local spec = self.spec_pipeBridge

    if self.isClient and self.firstTimeRun then
        if spec.tanker ~= nil and not self.isDeleted then
            -- Loop index issue.. avoid by manually calling the update for connection hoses.
            for index, hose in ipairs(self.spec_connectionHoses.updateableHoses) do
                if hose.connectedObject == spec.tanker and self.updateLoopIndex ~= hose.connectedObject.updateLoopIndex then
                    self:updateConnectionHose(hose, index)
                end
            end
        end
    end
end

function PipeBridge:onUpdateTick(dt)
    local spec = self.spec_pipeBridge

    if spec.forceHigherOnAttach then
        if self.isServer then
            local numOfAttachedImplements = 0
            if spec.attacherVehicle ~= nil then
                numOfAttachedImplements = #spec.attacherVehicle:getAttachedImplements()
            end

            if numOfAttachedImplements ~= spec.numOfAttachedImplements then
                spec.numOfAttachedImplementsSent = numOfAttachedImplements
                self:raiseDirtyFlags(spec.dirtyFlag)
            end
        end

        if spec.numOfAttachedImplementsSent ~= spec.numOfAttachedImplements then
            local upperAlpha = 1
            if spec.jointDesc ~= nil then
                upperAlpha = spec.jointDesc.upperAlpha
            end

            spec.heightTargetAlpha = upperAlpha
            spec.numOfAttachedImplements = spec.numOfAttachedImplementsSent

            if self:hasInputPipes() then
                if spec.tanker ~= nil then
                    local hoses = self:getConnectionHosesByInputAttacherJoint(self:getActiveInputAttacherJointDescIndex())
                    for _, hose in ipairs(hoses) do
                        self:disconnectHose(hose)
                    end
                    spec.tanker = nil
                end

                local tanker = PipeBridge.getAttachedPumpSourceObject(self)
                if tanker ~= nil then
                    local hoses = self:getConnectionHosesByInputAttacherJoint(self:getActiveInputAttacherJointDescIndex())
                    for _, hose in ipairs(hoses) do
                        self:disconnectHose(hose)
                    end
                    self:connectHosesToAttacherVehicle(tanker, 1, 1)
                    spec.tanker = tanker
                end
            end
        end
    end
end

function PipeBridge:loadPipesFromXML(xmlFile, baseKey, predicate)
    local pipes = {}

    local i = 0
    while true do
        local key = ("%s(%d)"):format(baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end

        local entry = {}
        if predicate(self, xmlFile, key, entry) then
            table.insert(pipes, entry)
        end

        i = i + 1
    end

    return pipes
end

function PipeBridge:loadOutputPipe(xmlFile, key, entry)
    entry.node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#node"), self.i3dMappings)

    if entry.node == nil then
        g_logManager:xmlWarning(self.configFileName, "Node not found!", key)
        return false
    end

    entry.scaleNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#scaleNode"), self.i3dMappings)

    if entry.scaleNode ~= nil then
        if not getHasShaderParameter(entry.scaleNode, "uvScale") then
            g_logManager:xmlWarning(self.configFileName, "Shader param uvScale not found '%s' for scale node!", key)
            return false
        end

        entry.maxLength = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#maxLength"), 3)
        entry.scaleRefNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#scaleRefNode"), self.i3dMappings)

        if entry.scaleRefNode == nil then
            g_logManager:xmlWarning(self.configFileName, "No scale ref not set!", key)
            return false
        end
    end

    entry.isLinked = false
    entry.orgTranslation = { getTranslation(entry.node) }

    return true
end

function PipeBridge:loadInputPipe(xmlFile, key, entry)
    entry.linkNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#linkNode"), self.i3dMappings)

    if entry.linkNode == nil then
        g_logManager:xmlWarning(self.configFileName, "LinkNode not found!", key)
        return false
    end

    entry.isLinked = false
    return true
end

function PipeBridge:hasOutputPipes()
    return #self.spec_pipeBridge.outputPipes ~= 0
end

function PipeBridge:hasInputPipes()
    return #self.spec_pipeBridge.inputPipes ~= 0
end

function PipeBridge:getOutputPipes()
    return self.spec_pipeBridge.outputPipes
end

function PipeBridge:getInputPipes()
    return self.spec_pipeBridge.inputPipes
end

function PipeBridge:bridgePipesWithVehicle(vehicle)
    local outputs = self:getOutputPipes()
    for i, input in ipairs(vehicle:getInputPipes()) do
        -- Not sure if index based is right... but works for now.
        if outputs[i] ~= nil then
            local output = outputs[i]
            link(input.linkNode, output.node)
            setTranslation(output.node, 0, 0, 0)

            output.isLinked = true
            output.linkedInput = input
            input.isLinked = true

            if output.scaleRefNode ~= nil then
                local px, py, pz = getWorldTranslation(output.scaleRefNode)
                local vx, vy, vz = getWorldTranslation(output.scaleNode)
                local distance = MathUtil.vector3Length(px - vx, py - vy, pz - vz)
                local _, y, z = worldToLocal(output.scaleNode, getWorldTranslation(output.scaleRefNode))
                local len = MathUtil.vector2Length(y, z)
                local scale = MathUtil.round(distance / len, 1)

                if scale > 1 then
                    scale = math.floor(scale + 0.5)
                end

                I3DUtil.setShaderParameterRec(output.scaleNode, "uvScale", 1, scale, 0, 0)
            end
        end
    end
end

function PipeBridge:canBridgePipesWithVehicle(vehicle)
    local outputs = self:getOutputPipes()
    for i, input in ipairs(vehicle:getInputPipes()) do
        -- Not sure if index based is right... but works for now.
        if outputs[i] ~= nil then
            local output = outputs[i]
            if output.scaleRefNode ~= nil then
                local px, py, pz = getWorldTranslation(input.linkNode)
                local vx, vy, vz = getWorldTranslation(output.scaleNode)
                local distance = MathUtil.round(MathUtil.vector3Length(px - vx, py - vy, pz - vz), 1)

                if distance > 1 then
                    distance = math.floor(distance + 0.5)
                end

                if distance > output.maxLength then
                    g_currentMission:showBlinkingWarning(g_i18n:getText("warning_requiresExtension"), 2000)
                    return false
                end
            end
        end
    end

    return true
end

function PipeBridge:unlinkPipesWithVehicle()
    local spec = self.spec_pipeBridge
    for i, output in ipairs(self:getOutputPipes()) do
        if output.isLinked then
            unlink(output.node)
            link(spec.rootNode, output.node)
            setTranslation(output.node, unpack(output.orgTranslation))
            output.isLinked = false

            if output.linkedInput ~= nil then
                output.linkedInput.isLinked = false
                output.linkedInput = nil
            end
        end
    end
end

function PipeBridge:getPipeBridgeAttacher(attacherVehicle)
    for _, implement in ipairs(attacherVehicle:getAttachedImplements()) do
        if implement.object ~= nil and implement.object ~= self then
            if SpecializationUtil.hasSpecialization(PipeBridge, implement.object.specializations) then
                return implement.object
            end
        end
    end

    return nil
end

function PipeBridge:updateAttacherJoint(moveAlpha)
    local spec = self.spec_pipeBridge
    local jointDesc = spec.jointDesc

    if jointDesc ~= nil then
        moveAlpha = self:setAttacherJointHeight(moveAlpha)

        local attacherVehicle = self:getAttacherVehicle()
        attacherVehicle:updateAttacherJointRotation(jointDesc, self)
        if self.isServer and jointDesc.jointIndex ~= 0 then
            setJointFrame(jointDesc.jointIndex, 0, jointDesc.jointTransform)
        end
    end

    spec.moveAlpha = moveAlpha
    self:raiseDirtyFlags(spec.dirtyFlag)
end

function PipeBridge:setAttacherJointHeight(moveAlpha)
    local spec = self.spec_pipeBridge
    local jointDesc = spec.jointDesc

    if moveAlpha == nil then
        moveAlpha = jointDesc.moveAlpha
    end

    moveAlpha = MathUtil.clamp(moveAlpha, jointDesc.upperAlpha, jointDesc.lowerAlpha)
    if jointDesc.rotationNode ~= nil then
        setRotation(jointDesc.rotationNode, MathUtil.vector3ArrayLerp(jointDesc.upperRotation, jointDesc.lowerRotation, moveAlpha))
    end
    if jointDesc.rotationNode2 ~= nil then
        setRotation(jointDesc.rotationNode2, MathUtil.vector3ArrayLerp(jointDesc.upperRotation2, jointDesc.lowerRotation2, moveAlpha))
    end

    return moveAlpha
end

function PipeBridge.getAttachedPumpSourceObject(object, rootObject, searchedImplements)
    if rootObject == nil then
        rootObject = object
    end

    if object ~= rootObject and SpecializationUtil.hasSpecialization(FillUnit, object.specializations) and not SpecializationUtil.hasSpecialization(PipeBridge, object.specializations) then
        local fillUnits = object:getFillUnits()
        for fillUnitIndex, _ in ipairs(fillUnits) do
            return object, fillUnitIndex
        end
    end

    if searchedImplements == nil then
        local attachedImplements
        if object.getAttacherVehicle ~= nil and object:getAttacherVehicle() ~= nil then
            local attacherVehicle = object:getAttacherVehicle()
            attachedImplements = attacherVehicle:getAttachedImplements()
        else
            attachedImplements = object:getAttachedImplements()
        end

        if attachedImplements ~= nil then
            for _, implement in ipairs(attachedImplements) do
                if implement.object ~= nil and implement.object ~= object then
                    local implementFound, fillUnitIndexFound = PipeBridge.getAttachedPumpSourceObject(implement.object, rootObject, true)

                    if implementFound ~= nil then
                        return implementFound, fillUnitIndexFound
                    end
                end
            end
        end
    end

    return nil
end

--- Events

function PipeBridge:onPostAttach(attacherVehicle, inputJointDescIndex, jointDescIndex)
    local bridgeVehicle = self:getPipeBridgeAttacher(attacherVehicle)
    local spec = self.spec_pipeBridge

    if bridgeVehicle ~= nil then
        local canConnect = false

        if self:hasOutputPipes() then
            canConnect = self:canBridgePipesWithVehicle(bridgeVehicle)
            if canConnect then
                self:bridgePipesWithVehicle(bridgeVehicle)
            end
        elseif bridgeVehicle:hasOutputPipes() then
            canConnect = bridgeVehicle:canBridgePipesWithVehicle(self)
            if canConnect then
                bridgeVehicle:bridgePipesWithVehicle(self)
            end
        end

        if canConnect and self:hasInputPipes() then
            local tanker = PipeBridge.getAttachedPumpSourceObject(self)
            if tanker ~= nil then
                self:connectHosesToAttacherVehicle(tanker, 1, 1)
                spec.tanker = tanker
            end
        end
    end

    spec.attacherVehicle = attacherVehicle

    if spec.forceHigherOnAttach then
        spec.jointDesc = attacherVehicle:getAttacherJointDescFromObject(self)

        ObjectChangeUtil.setObjectChanges(spec.jointDesc.changeObjects, false, self, self.setMovingToolDirty)

        spec.jointDesc.allowsLoweringBackup = spec.jointDesc.allowsLowering
        spec.jointDesc.allowsLowering = false

        spec.moveAlpha = self:setAttacherJointHeight()
        spec.heightTargetAlpha = spec.jointDesc.upperAlpha
    end
end

function PipeBridge:onPreDetach(attacherVehicle, implement)
    local spec = self.spec_pipeBridge

    if self:hasOutputPipes() then
        self:unlinkPipesWithVehicle()
    else
        local bridgeVehicle = self:getPipeBridgeAttacher(attacherVehicle)

        if bridgeVehicle ~= nil and bridgeVehicle:hasOutputPipes() then
            bridgeVehicle:unlinkPipesWithVehicle()
        end
    end

    if self:hasInputPipes() then
        if spec.tanker ~= nil then
            local hoses = self:getConnectionHosesByInputAttacherJoint(self:getActiveInputAttacherJointDescIndex())
            for _, hose in ipairs(hoses) do
                self:disconnectHose(hose)
            end
            spec.tanker = nil
        end
    end

    spec.attacherVehicle = nil

    if spec.jointDesc ~= nil then
        spec.jointDesc.allowsLowering = spec.jointDesc.allowsLoweringBackup
        spec.jointDesc = nil
    end
end

function PipeBridge:isDetachAllowed(superFunc)
    local spec = self.spec_pipeBridge

    -- When we are connected to a tanker we don't allow detach
    if spec.tanker ~= nil then
        return false, g_i18n:getText("info_detachTankerFirst"):format(spec.tanker.typeDesc)
    end

    if self:hasInputPipes() then
        for _, input in ipairs(self:getInputPipes()) do
            if input.isLinked then
                return false, nil
            end
        end
    end

    return superFunc(self)
end

function PipeBridge:canTurnOnPump(superFunc)
    if self:hasInputPipes() then
        for _, input in ipairs(self:getInputPipes()) do
            if not input.isLinked then
                return false
            end
        end
    end

    if self:hasOutputPipes() then
        for _, output in ipairs(self:getOutputPipes()) do
            if not output.isLinked then
                return false
            end
        end
    end

    return superFunc(self)
end
